# Copyright (c) 2023 Jan Malakhovski <oxij@oxij.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# This file is a part of pwebarc project.

import logging as _logging
import os as _os
import stat as _stat
import subprocess as _subprocess
import sys as _sys
import tempfile as _tempfile
import traceback as _traceback
import typing as _t

from gettext import gettext as _, ngettext

from kisstdlib import argparse
from kisstdlib.exceptions import *
from kisstdlib.io import *
from kisstdlib.logging import *

from .wrr import *
from .output import *

def is_filtered_out(rrexpr : ReqresExpr, cargs : _t.Any) -> bool:
    def eval_it(expr : str) -> bool:
        ev = rrexpr.eval(expr)
        if not isinstance(ev, bool):
            raise CatastrophicFailure("expression `%s` does not evaluate to `bool`, got `%s` instead", expr, repr(ev))
        return ev

    if len(cargs.anys) > 0:
        res = False
    else:
        res = True

    for expr in cargs.anys:
        res = res or eval_it(expr)

    for expr in cargs.alls:
        res = res and eval_it(expr)

    return not res

def cmd_pprint(cargs : _t.Any) -> None:
    def emit(reqres : Reqres, abs_path : str, rel_path : str) -> None:
        rrexpr = ReqresExpr(reqres, abs_path)
        if is_filtered_out(rrexpr, cargs): return

        wrr_pprint(stdout, reqres, abs_path, cargs.abridged)
        stdout.flush()

    for _ in wrr_map_paths(emit, cargs.paths, cargs.errors):
        pass

def get_bytes(expr : str, path : str) -> bytes:
    abs_path = _os.path.abspath(_os.path.expanduser(path))
    reqres = wrr_loadf(abs_path)
    rrexpr = ReqresExpr(reqres, abs_path)
    value = rrexpr.eval(expr)

    if value is None or isinstance(value, (bool, int, float)):
        value = str(value)
    elif isinstance(value, EpochMsec):
        value = str(int(value))

    if isinstance(value, str):
        return value.encode(_sys.getdefaultencoding())
    elif isinstance(value, bytes):
        return value
    else:
        raise Failure("don't know how to print an expression of type `%s`", type(value).__name__)

def cmd_get(cargs : _t.Any) -> None:
    data = get_bytes(cargs.expr, cargs.path)
    stdout.write_bytes(data)

def cmd_run(cargs : _t.Any) -> None:
    if cargs.num_args < 1:
        raise Failure("must have at least one PATH")
    elif cargs.num_args - 1 > len(cargs.args):
        raise Failure("not enough arguments to satisfy `--num-args`")

    # move (num_args - 1) arguments from args to paths
    ntail = len(cargs.args) + 1 - cargs.num_args
    args = cargs.args[:ntail]
    paths = cargs.args[ntail:] + cargs.paths

    tmp_paths = []
    try:
        for path in paths:
            data = get_bytes(cargs.expr, path)

            # TODO: extension guessing
            fileno, tmp_path = _tempfile.mkstemp(prefix = "wrrarms_run_", suffix = ".tmp")
            tmp_paths.append(tmp_path)

            with TIOWrappedWriter(_os.fdopen(fileno, "wb")) as f:
                f.write_bytes(data)

        retcode = _subprocess.Popen([cargs.command] + args + tmp_paths).wait()
        _sys.exit(retcode)
    finally:
        for path in tmp_paths:
            _os.unlink(path)

def slurp_stdin0(cargs : _t.Any) -> None:
    if not cargs.stdin0: return
    paths = stdin.read_all_bytes().split(b"\0")
    last = paths.pop()
    if last != b"":
        raise Failure("`--stdin0` input format error")
    cargs.paths += paths

def cmd_find(cargs : _t.Any) -> None:
    slurp_stdin0(cargs)

    def emit(reqres : Reqres, abs_path : str, rel_path : str) -> None:
        rrexpr = ReqresExpr(reqres, abs_path)
        if is_filtered_out(rrexpr, cargs): return
        stdout.write_bytes(_os.fsencode(abs_path) + cargs.terminator)
        stdout.flush()

    for _ in wrr_map_paths(emit, cargs.paths, cargs.errors):
        pass

output_aliases = {
    "default":  "%(ryear)d/%(rmonth)02d/%(rday)02d/%(rhour)02d%(rminute)02d%(rsecond)02d%(rtime_msq)03d_%(stime_ms)s_%(method)s_%(net_url|sha256|prefix 4)s_%(status)s_%(hostname)s.%(num)d.wrr",
    "short": "%(ryear)d/%(rmonth)02d/%(rday)02d/%(rtime_ms)d_%(stime_ms)s.%(num)d.wrr",
    "surl_msn":   "%(scheme)s/%(netloc)s/%(path|abbrev 120)s%(oqm)s%(query|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "url_msn":               "%(netloc)s/%(path|abbrev 120)s%(oqm)s%(query|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "shpq_msn":   "%(scheme)s/%(hostname)s/%(path|abbrev 120)s%(oqm)s%(query|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "hpq_msn":               "%(hostname)s/%(path|abbrev 120)s%(oqm)s%(query|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "shupq_msn":  "%(scheme)s/%(hostname)s/%(path|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "hupq_msn":              "%(hostname)s/%(path|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "srhupq_msn": "%(scheme)s/%(rhostname)s/%(path|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "rhupq_msn":             "%(rhostname)s/%(path|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 100)s_%(method)s_%(status)s.%(num)d.wrr",
    "full_shpq":  "%(scheme)s/%(hostname)s/%(ipath)s%(oqm)s%(query)s.wrr",
    "full_hpq":              "%(hostname)s/%(ipath)s%(oqm)s%(query)s.wrr",
    "shpq":       "%(scheme)s/%(hostname)s/%(ipath|abbrev 120)s%(oqm)s%(query|abbrev 120)s.wrr",
    "hpq":                   "%(hostname)s/%(ipath|abbrev 120)s%(oqm)s%(query|abbrev 120)s.wrr",
    "shupq":      "%(scheme)s/%(hostname)s/%(ipath|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 120)s.wrr",
    "hupq":                  "%(hostname)s/%(ipath|unquote|abbrev 120)s%(oqm)s%(query|unquote_plus|abbrev 120)s.wrr",
    "hupnq":                 "%(hostname)s/%(ipath|unquote|abbrev 120)s%(oqm)s%(nquery|unquote_plus|abbrev 120)s.wrr",
}

def make_organize_emit(cargs : _t.Any, destination : str) -> _t.Callable[[Reqres, str, str], None]:
    destination = _os.path.expanduser(destination)

    action_func : _t.Any
    if cargs.action == "rename":
        action_desc = "renaming"
        action_func = _os.rename
    elif cargs.action == "hardlink":
        action_desc = "hardlinking"
        action_func = _os.link
    elif cargs.action == "symlink":
        action_desc = "symlinking"
        action_func = _os.symlink
    elif cargs.action == "symlink-update":
        action_desc = "updating symlink"
        action_func = _os.symlink
    else:
        assert False

    seen_count_state : dict[str, int] = {}
    def seen_count(value : str) -> int:
        try:
            count = seen_count_state[value]
        except KeyError:
            seen_count_state[value] = 0
            return 0
        else:
            count += 1
            seen_count_state[value] = count
            return count

    last_updated_ms_state : dict[str, int] = {}

    def emit(reqres : Reqres, abs_path : str, rel_path : str) -> None:
        rrexpr = ReqresExpr(reqres, abs_path)
        if is_filtered_out(rrexpr, cargs): return

        rrexpr.items["num"] = 0
        ogprefix = _os.path.join(destination, cargs.output % rrexpr)
        prev_rel_out_path = None
        need_to_unlink = False
        while True:
            rrexpr.items["num"] = seen_count(ogprefix)
            rel_out_path = _os.path.join(destination, cargs.output % rrexpr)
            abs_out_path = _os.path.abspath(rel_out_path)
            if abs_path == abs_out_path:
                # trying to rename, hardlink, or symlink to itself
                return

            try:
                out_stat = _os.lstat(abs_out_path)
            except FileNotFoundError:
                break

            if _stat.S_ISLNK(out_stat.st_mode):
                # check that symlink target exists
                try:
                    out_stat_target = _os.stat(abs_out_path)
                except FileNotFoundError:
                    need_to_unlink = True
                    break

            if _os.path.samefile(abs_path, abs_out_path):
                # target already points to source
                return

            if cargs.action == "symlink-update":
                if not _stat.S_ISLNK(out_stat.st_mode):
                    raise Failure(f"trying to {cargs.action} `%s` to `%s` which already exists and is not a symlink; this is not allowed to prevent accidential data loss", rel_path, rel_out_path)

                this_update_ms = rrexpr.rtime_ms
                try:
                    last_update_ms = last_updated_ms_state[ogprefix]
                except KeyError:
                    last_update_ms = ReqresExpr(wrr_loadf(abs_out_path), abs_out_path).rtime_ms
                    last_updated_ms_state[ogprefix] = last_update_ms
                if last_update_ms >= this_update_ms:
                    # target in newer
                    return
                last_updated_ms_state[ogprefix] = this_update_ms

                need_to_unlink = True
                break

            if prev_rel_out_path == rel_out_path:
                raise Failure(f"trying to {cargs.action} `%s` to `%s` which already exists and is not the same file; meanwhile `--output` does not appear to allow for variance (did your forget to place a `%%(num)d` substitution in there?); this is not allowed to prevent accidential data loss", rel_path, rel_out_path)
            prev_rel_out_path = rel_out_path
            continue

        if cargs.dry_run:
            stderr.write_str_ln(f"dry-run: (not) {action_desc}: {rel_path} -> {rel_out_path}")
            stderr.flush()
            return

        if need_to_unlink:
            _os.unlink(abs_out_path)
        else:
            dirname = _os.path.dirname(abs_out_path)
            _os.makedirs(dirname, exist_ok = True)
        action_func(abs_path, abs_out_path)

        stderr.write_str_ln(f"{action_desc}: {rel_path} -> {rel_out_path}")
        stderr.flush()
        if cargs.terminator is not None:
            stdout.write_bytes(_os.fsencode(abs_out_path) + cargs.terminator)
            stdout.flush()

    return emit

def cmd_organize(cargs : _t.Any) -> None:
    if cargs.output in output_aliases:
        cargs.output = output_aliases[cargs.output]

    if cargs.destination is not None:
        # destination is set explicitly
        slurp_stdin0(cargs)

        emit = make_organize_emit(cargs, cargs.destination)
        for _ in wrr_map_paths(emit, cargs.paths, cargs.errors, follow_symlinks=False):
            pass
    else:
        # each path is its own destination
        if cargs.stdin0:
            raise Failure("`--stdin0` but no `--to` is specified")

        for path in cargs.paths:
            try:
                fstat = _os.stat(_os.path.expanduser(path))
            except FileNotFoundError:
                raise Failure("%s does not exist", path)
            else:
                if cargs.destination is None and not _stat.S_ISDIR(fstat.st_mode):
                    raise Failure("%s is not a directory but no `--to` is specified", path)

        for path in cargs.paths:
            emit = make_organize_emit(cargs, path)
            for _ in wrr_map_paths(emit, [path], cargs.errors, follow_symlinks=False):
                pass

def get_StreamEncoder(cargs : _t.Any) -> StreamEncoder:
    stream : StreamEncoder
    if cargs.format == "py":
        stream = PyStreamEncoder(stdout, cargs.abridged)
    elif cargs.format == "cbor":
        stream = CBORStreamEncoder(stdout, cargs.abridged)
    elif cargs.format == "json":
        stream = JSONStreamEncoder(stdout, cargs.abridged)
    elif cargs.format == "raw":
        stream = RawStreamEncoder(stdout, cargs.abridged, cargs.terminator)
    else:
        assert False
    return stream

def cmd_stream(cargs : _t.Any) -> None:
    stream = get_StreamEncoder(cargs)

    def emit(reqres : Reqres, abs_path : str, rel_path : str) -> None:
        rrexpr = ReqresExpr(reqres, abs_path)
        if is_filtered_out(rrexpr, cargs): return

        values : list[_t.Any] = []
        for expr in cargs.exprs:
            values.append(rrexpr.eval(expr))
        stream.emit(abs_path, cargs.exprs, values)

    stream.start()
    for _ in wrr_map_paths(emit, cargs.paths, cargs.errors):
        pass
    stream.finish()

def add_doc(fmt : argparse.BetterHelpFormatter) -> None:
    fmt.add_text(_("# Examples"))

    fmt.start_section(_("Pretty-print all reqres in `../dumb_server/pwebarc-dump` using an abridged (for ease of reading and rendering) verbose textual representation"))
    fmt.add_code(f"{__package__} pprint ../dumb_server/pwebarc-dump")
    fmt.end_section()

    fmt.start_section(_("Pipe response body from a given WRR file to stdout"))
    fmt.add_code(f"{__package__} get ../dumb_server/pwebarc-dump/path/to/file.wrr")
    fmt.end_section()

    fmt.start_section(_("Get first 4 characters of a hex digest of sha256 hash computed on the URL without the fragment/hash part"))
    fmt.add_code(f'{__package__} get -e "net_url|sha256|prefix 4" ../dumb_server/pwebarc-dump/path/to/file.wrr')
    fmt.end_section()

    fmt.start_section(_("Pipe response body from a given WRR file to stdout, but less efficiently, by generating a temporary file and giving it to `cat`"))
    fmt.add_code(f"{__package__} run cat ../dumb_server/pwebarc-dump/path/to/file.wrr")
    fmt.add_text(_(f"Thus `{__package__} run` can be used to do almost anything you want, e.g."))
    fmt.add_code(f"{__package__} run less ../dumb_server/pwebarc-dump/path/to/file.wrr")
    fmt.add_code(f"{__package__} run -- sort -R ../dumb_server/pwebarc-dump/path/to/file.wrr")
    fmt.add_code(f"{__package__} run -n 2 -- diff -u ../dumb_server/pwebarc-dump/path/to/file-v1.wrr ../dumb_server/pwebarc-dump/path/to/file-v2.wrr")
    fmt.end_section()

    fmt.start_section(_(f"List paths of all WRR files from `../dumb_server/pwebarc-dump` that contain only complete `200 OK` responses with bodies larger than 1K"))
    fmt.add_code(f"""wrrarms find --and "status|== 200C" --and "response.body|len|> 1024" ../dumb_server/pwebarc-dump""")
    fmt.end_section()

    fmt.start_section(_(f"Rename all WRR files in `../dumb_server/pwebarc-dump/default` according to their metadata using `--output default` (see the `{__package__} organize` section for its definition, the `default` format is designed to be human-readable while causing almost no collisions, thus making `num` substitution parameter to almost always stay equal to `0`, making things nice and deterministic)"))
    fmt.add_code(f"{__package__} organize ../dumb_server/pwebarc-dump/default")
    fmt.add_text(_("alternatively, just show what would be done"))
    fmt.add_code(f"{__package__} organize --dry-run ../dumb_server/pwebarc-dump/default")
    fmt.end_section()

    fmt.start_section(_(f"The output of `{__package__} organize --zero-terminated` can be piped into `{__package__} organize --stdin0` to perform complex updates. E.g. the following will rename new reqres from `../dumb_server/pwebarc-dump` to `~/pwebarc/raw` renaming them with `--output default`, the `for` loop is there to preserve profiles"))
    fmt.add_code(f"""for arg in ../dumb_server/pwebarc-dump/* ; do
  wrrarms organize --zero-terminated --to ~/pwebarc/raw/"$(basename "$arg")" "$arg"
done > changes""")
    fmt.add_text(_("then, we can reuse `changes` to symlink all new files from `~/pwebarc/raw` to `~/pwebarc/all` using `--output hupq_msn`, which would show most of the URL in the file name:"))
    fmt.add_code(f"""wrrarms organize --stdin0 --action symlink --to ~/pwebarc/all --output hupq_msn < changes""")
    fmt.add_text(_("and then, we can reuse `changes` again and use them to update `~/pwebarc/latest`, filling it with symlinks pointing to the latest `200 OK` complete reqres from `~/pwebarc/raw`, similar to what `wget -r` would produce (except `wget` would do network requests and produce responce bodies, while this will build a file system tree of symlinks to WRR files in `/pwebarc/raw`):"))
    fmt.add_code(f"""wrrarms organize --stdin0 --action symlink-update --to ~/pwebarc/latest --output hupq --and "status|== 200C" < changes""")
    fmt.end_section()

    fmt.add_text(_("# Advanced examples"))

    fmt.start_section(_("Pretty-print all reqres in `../dumb_server/pwebarc-dump` by dumping their whole structure into an abridged Pythonic Object Representation (repr)"))
    fmt.add_code(f"{__package__} stream --expr . ../dumb_server/pwebarc-dump")
    fmt.add_code(f"{__package__} stream -e . ../dumb_server/pwebarc-dump")
    fmt.end_section()

    fmt.start_section(_("Pretty-print all reqres in `../dumb_server/pwebarc-dump` using the unabridged verbose textual representation"))
    fmt.add_code(f"{__package__} pprint --unabridged ../dumb_server/pwebarc-dump")
    fmt.add_code(f"{__package__} pprint -u ../dumb_server/pwebarc-dump")
    fmt.end_section()

    fmt.start_section(_("Pretty-print all reqres in `../dumb_server/pwebarc-dump` by dumping their whole structure into the unabridged Pythonic Object Representation (repr) format"))
    fmt.add_code(f"{__package__} stream --unabridged --expr . ../dumb_server/pwebarc-dump")
    fmt.add_code(f"{__package__} stream -ue . ../dumb_server/pwebarc-dump")
    fmt.end_section()

    fmt.start_section(_("Produce a JSON list of `[<file path>, <time it finished loading in milliseconds since UNIX epoch>, <URL>]` tuples (one per reqres) and pipe it into `jq` for indented and colored output"))
    fmt.add_code(f"{__package__} stream --format=json -ue fs_path -e finished_at -e request.url ../dumb_server/pwebarc-dump | jq .")
    fmt.end_section()

    fmt.start_section(_("Similarly, but produce a CBOR output"))
    fmt.add_code(f"{__package__} stream --format=cbor -ue fs_path -e finished_at -e request.url ../dumb_server/pwebarc-dump | less")
    fmt.end_section()

    fmt.start_section(_("Concatenate all response bodies of all the requests in `../dumb_server/pwebarc-dump`"))
    fmt.add_code(f'{__package__} stream --format=raw --not-terminated -ue "response.body|es" ../dumb_server/pwebarc-dump | less')
    fmt.end_section()

    fmt.start_section(_("Print all unique visited URLs, one per line"))
    fmt.add_code(f"{__package__} stream --format=raw --lf-terminated -ue request.url ../dumb_server/pwebarc-dump | sort | uniq")
    fmt.end_section()

    fmt.start_section(_("Same idea, but using NUL bytes while processing, and prints two URLs per line"))
    fmt.add_code(f"{__package__} stream --format=raw --zero-terminated -ue request.url ../dumb_server/pwebarc-dump | sort -z | uniq -z | xargs -0 -n2 echo")
    fmt.end_section()

    fmt.add_text(_("# Handling binary data"))

    fmt.add_text(_(f"Trying to use response bodies produced by `{__package__} stream --format=json` is likely to result garbled data as JSON can't represent raw sequences of bytes, thus binary data will have to be encoded into UNICODE using replacement characters:"))
    fmt.add_code(f"{__package__} stream --format=json -ue . ../dumb_server/pwebarc-dump/path/to/file.wrr | jq .")
    fmt.add_text(_("The most generic solution to this is to use `--format=cbor` instead, which would produce a verbose CBOR representation equivalent to the one used by `--format=json` but with binary data preserved as-is:"))
    fmt.add_code(f"{__package__} stream --format=cbor -ue . ../dumb_server/pwebarc-dump/path/to/file.wrr | less")
    fmt.add_text(_("Or you could just dump raw response bodies separately:"))
    fmt.add_code(f"{__package__} stream --format=raw -ue response.body ../dumb_server/pwebarc-dump/path/to/file.wrr | less")
    fmt.add_code(f"{__package__} get ../dumb_server/pwebarc-dump/path/to/file.wrr | less")

def main() -> None:
    parser = argparse.BetterArgumentParser(
        prog=__package__,
        description=_("A tool to pretty-print, compute and print values from, search, organize (programmatically rename/move/symlink/hardlink files), (WIP: check, deduplicate, and edit) pWebArc WRR (WEBREQRES, Web REQuest+RESponse) archive files.") + "\n\n" +
_("Terminology: a `reqres` (`Reqres` when a Python type) is an instance of a structure representing HTTP request+response pair with some additional metadata."),
        additional_sections = [add_doc],
        allow_abbrev = False,
        add_version = True,
        add_help = False)
    parser.add_argument("-h", "--help", action="store_true", help=_("show this help message and exit"))
    parser.add_argument("--markdown", action="store_true", help=_("show help messages formatted in Markdown"))

    subparsers = parser.add_subparsers(title="subcommands")

    def no_cmd(cargs : _t.Any) -> None:
        parser.print_help(stderr) # type: ignore
        _sys.exit(2)
    parser.set_defaults(func=no_cmd)

    def add_errors(cmd : _t.Any) -> None:
        cmd.add_argument("--errors", choices=["fail", "skip", "ignore"], default="fail", help=_("""when an error occurs:
- `fail`: report failure and stop the execution (default)
- `skip`: report failure but skip the reqres that produced it from the output and continue
- `ignore`: `skip`, but don't report the failure"""))

    def add_filters(cmd : _t.Any) -> None:
        grp = cmd.add_argument_group("filters")
        grp.add_argument("--or", dest="anys", metavar="EXPR", action="append", type=str, default = [],
                         help=_(f"only work on reqres which match any of these expressions..."))
        grp.add_argument("--and", dest="alls", metavar="EXPR", action="append", type=str, default = [],
                         help=_(f"... and all of these expressions, both can be specified multiple times, both use the same expression format as `{__package__} get --expr`, which see"))

    def add_abridged(cmd : _t.Any) -> None:
        grp = cmd.add_mutually_exclusive_group()
        grp.add_argument("-u", "--unabridged", dest="abridged", action="store_false", help=_("print all data in full"))
        grp.add_argument("--abridged", action="store_true", help=_("shorten long strings for brevity (useful when you want to visually scan through batch data dumps) (default)"))
        cmd.set_defaults(abridged = True)

    def add_paths(cmd : _t.Any) -> None:
        cmd.add_argument("paths", metavar="PATH", nargs="*", type=str, help=_("inputs, can be a mix of files and directories (which will be traversed recursively)"))

    # pprint
    cmd = subparsers.add_parser("pprint", help=_("pretty-print WRR files"),
                                description = _("""Pretty-print given WRR files to stdout."""))
    add_errors(cmd)
    add_filters(cmd)
    add_abridged(cmd)
    add_paths(cmd)
    cmd.set_defaults(func=cmd_pprint)

    # get
    cmd = subparsers.add_parser("get", help=_("print an expression computed from a WRR file to stdout"),
                                description = _(f"""Compute an expression EXPR for a reqres stored at PATH and then print it to stdout."""))
    cmd.add_argument("-e", "--expr", metavar="EXPR", default = "response.body|es", help=_('an expression to compute (default: `%(default)s`), a state-transformer (pipeline) which starts from value `None` and applies to it a program built from the following:') + "\n" + \
                     "- " + _("constants and functions:") + "\n" + \
                     "".join([f"  - `{name}`: {_(value[0]).replace('%', '%%')}\n" for name, value in linst_atoms.items()]) + \
                     "- " + _("reqres fields, these work the same way as constants above, i.e. they replace current value of `None` with field's value, if reqres is missing the field in question, which could happen for `response*` fields, the result is `None`:") + "\n" + \
                     "".join([f"  - `{name}`: {_(value).replace('%', '%%')}\n" for name, value in Reqres_fields.items()]) + \
                     "- " + _("derived attributes:") + "\n" + \
                     "".join([f"  - `{name}`: {_(value).replace('%', '%%')}\n" for name, value in Reqres_derived_attrs.items()]) + \
                     "- " + _("a compound expression built by piping (`|`) the above, for example:") + """
  - `net_url|sha256`
  - `net_url|sha256|prefix 4`
  - `path|unquote`
  - `query|unquote_plus|abbrev 128`
  - `response.complete`: this will print the value of `response.complete` or `None`, if there was no response
  - `response.complete|false`: this will print `response.complete` or `False`
  - `response.body|eb`: this will print `response.body` or an empty string, if there was no response
""")

    cmd.add_argument("path", metavar="PATH", type=str, help=_("input WRR file path"))
    cmd.set_defaults(func=cmd_get)

    # run
    cmd = subparsers.add_parser("run", help=_("spawn a process on generated temporary files produced from expressions computed on WRR files"),
                                description = _("""Compute an expression EXPR for each of NUM reqres stored at PATHs, dump the results into into newly created temporary files, spawn a given COMMAND with given arguments ARGs and the resulting temporary file paths appended as the last NUM arguments, wait for it to finish, delete the temporary files, exit with the return code of the spawned process."""))
    cmd.add_argument("-e", "--expr", metavar="EXPR", default = "response.body|es", help=_("the expression to compute, see `{__package__} get --expr` for more info  on expression format (default: `%(default)s`)"))
    cmd.add_argument("-n", "--num-args", metavar="NUM", type=int, default = 1, help=_("number of PATHs (default: `%(default)s`)"))
    cmd.add_argument("command", metavar="COMMAND", type=str, help=_("command to spawn"))
    cmd.add_argument("args", metavar="ARG", nargs="*", type=str, help=_("additional arguments to give to the COMMAND"))
    cmd.add_argument("paths", metavar="PATH", nargs="+", type=str, help=_("input WRR file paths to be mapped into new temporary files"))
    cmd.set_defaults(func=cmd_run)

    def add_stdin0(cmd : _t.Any) -> None:
        cmd.add_argument("--stdin0", action="store_true", help=_("read zero-terminated PATHs from stdin, these will be processed after PATHs specified as command-line arguments, requires specified `--to`"))

    # find
    cmd = subparsers.add_parser("find", help=_("print paths of WRR files matching specified criteria"),
                                description = _(f"""Print paths of WRR files matching specified criteria."""))
    add_errors(cmd)
    add_filters(cmd)
    add_stdin0(cmd)

    grp = cmd.add_mutually_exclusive_group()
    grp.add_argument("-l", "--lf-terminated", dest="terminator", action="store_const", const = b"\n", help=_("output absolute paths of matching WRR files terminated with `\\n` (LF) newline characters to stdout (default)"))
    cmd.add_argument("-z", "--zero-terminated", dest="terminator", action="store_const", const = b"\0", help=_("output absolute paths of matching WRR files terminated with `\\0` (NUL) bytes to stdout"))
    cmd.set_defaults(terminator = b"\n")

    add_paths(cmd)
    cmd.set_defaults(func=cmd_find)

    # organize
    cmd = subparsers.add_parser("organize", help=_("rename/hardlink/symlink WRR files based on their metadata"),
                                description = _(f"""Rename/hardlink/symlink given WRR files to DESTINATION based on their metadata.

Operations that could lead to accidental data loss are not permitted.
E.g. `{__package__} organize --action rename` will not overwrite any files, which is why the default `--output` contains `%(num)d`."""))
    add_errors(cmd)
    add_filters(cmd)
    cmd.add_argument("--dry-run", action="store_true", help=_("perform a trial run without actually performing any changes"))
    add_stdin0(cmd)

    grp = cmd.add_mutually_exclusive_group()
    grp.add_argument("-n", "--no-output", dest="terminator", action="store_const", const = None, help=_("don't print anything to stdout (default)"))
    grp.add_argument("-l", "--lf-terminated", dest="terminator", action="store_const", const = b"\n", help=_("output absolute paths of newly produced files terminated with `\\n` (LF) newline characters to stdout"))
    cmd.add_argument("-z", "--zero-terminated", dest="terminator", action="store_const", const = b"\0", help=_("output absolute paths of newly produced files terminated with `\\0` (NUL) bytes to stdout"))
    cmd.set_defaults(terminator = None)

    cmd.add_argument("--action", choices=["rename", "hardlink", "symlink", "symlink-update"], default="rename", help=_("""organize how:
- `rename`: rename source files under DESTINATION, will fail if target already exists (default)
- `hardlink`: create hardlinks from source files to paths under DESTINATION, will fail if target already exists
- `symlink`: create symlinks from source files to paths under DESTINATION, will fail if target already exists
- `symlink-update`: create symlinks from source files to paths under DESTINATION, will overwrite the target if `rtime_ms` for the source reqres is newer than the same value for the target
"""))
    cmd.add_argument("-o", "--output", metavar="FORMAT", default="default", type=str, help=_("format describing the generated output path, an alias name or a custom pythonic %%-substitution string:") + "\n" + \
                     "- " + _("available aliases and corresponding %%-substitutions:") + "\n" + \
                     "".join([f"  - `{name}`: `{value.replace('%', '%%')}`" + (" (default)" if name == "default" else "") + "\n" for name, value in output_aliases.items()]) + \
                     "- " + _("available substitutions:") + "\n" + \
                     "  - `num`: " + _("number of times an output path like this was seen; this value gets incremened for each new WRR file generating the same path with `num` set to `0` and when the file at the path generated with the current value of `num` already exists; i.e. adding this parameter to your `--output` format will ensure all generated file names will be unique") + "\n" + \
                     "  - " + _(f"all expressions of `{__package__} get --expr`, which see"))
    cmd.add_argument("-t", "--to", dest="destination", metavar="DESTINATION", type=str, help=_("target directory, if not set then each source PATH must be a directory which will be is its own DESTINATION"))
    add_paths(cmd)
    cmd.set_defaults(func=cmd_organize)

    # stream
    cmd = subparsers.add_parser("stream", help=_(f"produce a stream structured lists containing expressions computed from specified WRR files to stdout, a generalized `{__package__} get`"),
                                description = _("""Compute given expressions for each of given WRR files, encode them into a requested format, and print the result to stdout."""))
    add_errors(cmd)
    add_filters(cmd)
    add_abridged(cmd)

    cmd.add_argument("--format", choices=["py", "cbor", "json", "raw"], default="py", help=_("""generate output in:
- py: Pythonic Object Representation aka `repr` (default)
- cbor: CBOR (RFC8949)
- json: JavaScript Object Notation aka JSON (binary data can't be represented, UNICODE replacement characters will be used)
- raw: concatenate raw values (termination is controlled by `*-terminated` options)
"""))

    grp = cmd.add_mutually_exclusive_group()
    grp.add_argument("-l", "--lf-terminated", dest="terminator", action="store_const", const = b"\n", help=_("terminate `raw` output values with `\\n` (LF) newline characters (default)"))
    cmd.add_argument("-z", "--zero-terminated", dest="terminator", action="store_const", const = b"\0", help=_("terminate `raw` output values with `\\0` (NUL) bytes"))
    grp.add_argument("-n", "--not-terminated", dest="terminator", action="store_const", const = b"", help=_("don't terminate `raw` output values with anything, just concatenate them"))
    cmd.set_defaults(terminator = b"\n")

    cmd.add_argument("-e", "--expr", dest="exprs", metavar="EXPR", action="append", type=str, default = [], help=_(f'an expression to compute, see `{__package__} get --expr` for more info on expression format, can be specified multiple times (default: `%(default)s`); to dump all the fields of a reqres, specify "`.`"'))
    add_paths(cmd)
    cmd.set_defaults(func=cmd_stream)

    cargs = parser.parse_args(_sys.argv[1:])

    if cargs.help:
        if cargs.markdown:
            parser.set_formatter_class(argparse.MarkdownBetterHelpFormatter)
            print(parser.format_help(1024))
        else:
            print(parser.format_help())
        _sys.exit(0)

    _logging.basicConfig(level=_logging.WARNING,
                         stream = stderr)
    errorcnt = CounterHandler()
    logger = _logging.getLogger()
    logger.addHandler(errorcnt)
    #logger.setLevel(_logging.DEBUG)

    try:
        cargs.func(cargs)
    except KeyboardInterrupt:
        stderr.write_str_ln("Interrupted!")
        errorcnt.errors += 1
    except CatastrophicFailure as exc:
        stderr.write_str_ln("error: " + str(exc))
        errorcnt.errors += 1
    except Exception as exc:
        _traceback.print_exception(type(exc), exc, exc.__traceback__, 100, stderr)
        errorcnt.errors += 1

    stdout.flush()
    stderr.flush()

    if errorcnt.errors > 0:
        stderr.write_str_ln(ngettext("There was %d error!", "There were %d errors!", errorcnt.errors) % (errorcnt.errors,))
        _sys.exit(1)
    _sys.exit(0)

if __name__ == "__main__":
    main()



# Copyright (c) 2023 Jan Malakhovski <oxij@oxij.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of pwebarc project.

import cbor2.decoder as _cbor2dec
import cbor2.encoder as _cbor2enc
import io as _io
import json as _json
import typing as _t

from kisstdlib.exceptions import *
from kisstdlib.io import *

from .wrr import *
from .type import *

def abridge_anystr(value : _t.AnyStr, length : int, ln : bool) -> tuple[bool, _t.AnyStr]:
    hlength = length // 2
    if len(value) > length:
        if isinstance(value, bytes):
            return True, value[:hlength] + (b"\n...\n" if ln else b" ... ") + value[-hlength:]
        else:
            return True, value[:hlength] + ("\n...\n" if ln else " ... ") + value[-hlength:]
    else:
        return False, value

def wrr_pprint(fobj : TIOWrappedWriter, reqres : Reqres, path : str | bytes, abridge : bool) -> None:
    req = reqres.request
    res = reqres.response

    code = "?"
    reason = ""
    if res is not None:
        code = str(res.code)
        reason = res.reason

    if path is not None:
        fobj.write_str("file ")
        if isinstance(path, str):
            fobj.write_str_ln(path)
        elif isinstance(path, bytes):
            fobj.write_bytes_ln(path)
        else:
            assert False
    fobj.write_str_ln(f"WEBREQRES/{str(reqres.version)} {reqres.protocol} {req.method} {req.url} {code} {reason}")
    fobj.write_str_ln(f"source {reqres.source}")

    req_complete = "incomplete"
    if req.complete:
        req_complete = "complete"
    fobj.write_str_ln(f"request {req_complete} {str(len(req.headers.value))} headers {str(len(req.body))} bytes")

    if res is not None:
        res_complete = "incomplete"
        if res.complete:
            res_complete = "complete"
        fobj.write_str_ln(f"response {res_complete} {str(len(res.headers.value))} headers {str(len(res.body))} bytes")
    else:
        fobj.write_str_ln("response none")

    fobj.write_str_ln(f"clock {fmt_msec_interval(req.started_at, reqres.finished_at)}")

    if len(reqres.extra) > 0:
        for k, v in reqres.extra.items():
            fobj.write_str(k + " ")
            fobj.write_bytes(pyrepr_dumps(v, starting_indent = 2, width = 80, default = PyStreamEncoder.encode_py).lstrip())
        fobj.write_str_ln("")

    def dump_data(data : str | bytes, complete : str) -> None:
        def encode(value : str) -> bytes:
            return value.encode(fobj.encoding)

        final = False
        raw = False
        if isinstance(data, str):
            if data.find("\0") != -1:
                guess = "binary"
            else:
                guess = f"text encoding={fobj.encoding}"
                raw = True
            data = encode(data)
        elif isinstance(data, bytes):
            guess = "binary"
        else:
            assert False

        if data.startswith(b"\x89PNG"):
            guess = "image/png"
            final, raw = True, False
        elif data.startswith(b"\xff\xd8\xff\xe0\x00\x10JFIF"):
            guess = "image/jpeg"
            final, raw = True, False
        elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
            guess = "image/webp"
            final, raw = True, False

        if not final:
            try:
                js = _json.loads(data.decode(fobj.encoding, errors = "replace"))
            except _json.decoder.JSONDecodeError:
                pass
            else:
                guess = f"json"
                jsdump = _json.dumps(js, ensure_ascii = False, indent = 2)
                data = encode(jsdump)
                final, raw = True, True

        if not final:
            try:
                with _io.BytesIO(data) as fp:
                    cb = _cbor2dec.CBORDecoder(fp).decode()
                    if len(fp.read()) != 0:
                        # not all of the data was decoded
                        raise Exception("failed to parse")
                cbordump = pyrepr_dumps(cb, width = 80)
            except Exception:
                pass
            else:
                guess = "cbor"
                data = cbordump
                final, raw = True, True

        if not final and guess.startswith("text"):
            data_strip = data.strip()
            if data_strip.startswith(b"<svg ") and data_strip.endswith(b"</svg>"):
                guess = "image/svg+xml"
                final, raw = True, True

        if not final and guess.startswith("text"):
            if data.find(b"<html") != -1:
                guess += "/html"
                if data.find(b"</html>") == -1:
                    guess += " incomplete=true"
                final, raw = True, True

        status = ""
        if raw:
            status += ", raw"
        else:
            data = encode(repr(data)[2:-1])
            status += ", quoted"

        abridged = abridge
        if abridge:
            abridged, data = abridge_anystr(data, 1024, True)
        if abridged:
            status += ", abridged"

        fobj.write_str_ln(f"({complete}, {len(data)} bytes, {guess}{status}):")
        fobj.write_bytes_ln(data)

    def dump_headers(headers : Headers, indent : str = "") -> None:
        for name, value in headers.value:
            fobj.write_str(indent + name + ": ")
            fobj.write_bytes_ln(value)

    fobj.write_str_ln("\nRequest headers:")
    dump_headers(req.headers, "  ")

    if res is not None:
        fobj.write_str_ln("\nResponse headers:")
        dump_headers(res.headers, "  ")

    if len(req.body) > 0:
        fobj.write_str("\nRequest body ")
        dump_data(req.body, req_complete)
    else:
        fobj.write_str_ln("\nEmpty request body")

    if res is not None:
        if len(res.body) > 0:
            fobj.write_str("\nResponse body ")
            dump_data(res.body, res_complete)
        else:
            fobj.write_str_ln("\nEmpty response body")

    fobj.write_str_ln("")

StreamEncoderElem = _t.TypeVar("StreamEncoderElem")
class StreamEncoder:
    def __init__(self, fobj : TIOWrappedWriter, abridged : bool) -> None:
        self.fobj = fobj
        self.abridged = abridged
        self.not_first = False

    def start(self) -> None:
        self.not_first = False

    def emit(self, path : str, names : list[str], values : list[_t.Any]) -> None:
        pass

    def finish(self) -> None:
        self.fobj.flush()

class CBORStreamEncoder(StreamEncoder):
    def __init__(self, fobj : TIOWrappedWriter, abridged : bool) -> None:
        super().__init__(fobj, abridged)

        encoders = _cbor2enc.default_encoders.copy()
        if abridged:
            encoders[bytes] = _t.cast(_t.Any, self.encode_cbor_abridged)
            encoders[str] = _t.cast(_t.Any, self.encode_cbor_abridged)
        self.encoder = _cbor2enc.CBOREncoder(_io.BytesIO(), default = self.encode_cbor)
        self.encoder._encoders = encoders

    @staticmethod
    def encode_cbor(enc : _cbor2enc.CBOREncoder, obj : _t.Any) -> None:
        enc.encode(plainify(obj))

    @staticmethod
    def encode_cbor_abridged(enc : _cbor2enc.CBOREncoder, obj : _t.Any) -> None:
        abridged, value = abridge_anystr(obj, 256, False)
        if isinstance(value, bytes):
            enc.encode_bytestring(value)
        elif isinstance(value, str):
            enc.encode_string(value)
        else:
            assert False

    def start(self) -> None:
        super().start()
        self.fobj.write_bytes(b"\x9f") # start indefinite-length array

    def emit(self, path : str, names : list[str], values : list[_t.Any]) -> None:
        try:
            self.encoder.encode(values)
            self.fobj.write_bytes(self.encoder.fp.getvalue())
        finally:
            self.encoder.fp = _io.BytesIO()

    def finish(self) -> None:
        self.fobj.write_bytes(b"\xff") # break symbol
        super().finish()

class PyStreamEncoder(StreamEncoder):
    def __init__(self, fobj : TIOWrappedWriter, abridged : bool) -> None:
        super().__init__(fobj, abridged)

        encoders = pyrepr_default_encoders.copy()
        if abridged:
            encoders[bytes] = _t.cast(_t.Any, self.encode_py_abridged)
            encoders[str] = _t.cast(_t.Any, self.encode_py_abridged)
        self.encoder = PyReprEncoder(_io.BytesIO(),
                                     width = 80, indent = 2, starting_indent = 0,
                                     encoders = encoders, default = self.encode_py)

    @staticmethod
    def encode_py(enc : PyReprEncoder, obj : _t.Any) -> None:
        if isinstance(obj, EpochMsec):
            enc.lexeme(str(int(obj)))
            enc.comment(str(obj))
        else:
            enc.encode(plainify(obj))

    @staticmethod
    def encode_py_abridged(enc : PyReprEncoder, obj : _t.AnyStr) -> None:
        abridged, value = abridge_anystr(repr(obj), 256, False)
        enc.lexeme(value)
        if abridged:
            if isinstance(obj, bytes):
                enc.comment(f"{len(obj)} bytes total")
            else:
                enc.comment(f"{len(obj)} characters total")

    def start(self) -> None:
        super().start()
        self.fobj.write_str_ln("[")

    def emit(self, path : str, names : list[str], values : list[_t.Any]) -> None:
        try:
            if self.not_first:
                self.encoder.write_str(",")
                self.encoder.flush_line(True)
            else:
                self.not_first = True
            self.encoder.encode(values)
            self.fobj.write_bytes(self.encoder.fobj.getvalue())
        finally:
            self.encoder.fobj = _io.BytesIO()

    def finish(self) -> None:
        self.fobj.write_str_ln("\n]")
        super().finish()

class JSONStreamEncoder(StreamEncoder):
    def start(self) -> None:
        super().start()
        self.fobj.write_str_ln("[")

    @staticmethod
    def encode_json(obj : _t.Any) -> _t.Any:
        if isinstance(obj, bytes):
            return obj.decode("utf-8", errors = "replace")
        return plainify(obj)

    @staticmethod
    def abridge_json(obj : _t.Any) -> _t.Any:
        try:
            obj = plainify(obj)
        except Exception:
            pass

        if obj is None or isinstance(obj, (bool, int, float)):
            return obj
        elif isinstance(obj, (str, bytes)):
            _, value = abridge_anystr(obj, 256, False) # type: ignore
            return value
        elif isinstance(obj, list):
            return [JSONStreamEncoder.abridge_json(a) for a in obj]
        elif isinstance(obj, dict):
            return {k: JSONStreamEncoder.abridge_json(v) for k, v in obj.items()}
        raise Failure("can't abridge a value of type `%s`", type(obj).__name__)

    def emit(self, path : str, names : list[str], values : list[_t.Any]) -> None:
        if self.abridged:
            values = self.abridge_json(values)
        data = _json.dumps(values, ensure_ascii = False, indent = 2, default = self.encode_json)
        if self.not_first:
            self.fobj.write_str_ln(",")
        else:
            self.not_first = True
        self.fobj.write_str(data)

    def finish(self) -> None:
        self.fobj.write_str_ln("\n]")
        super().finish()

class RawStreamEncoder(StreamEncoder):
    def __init__(self, fobj : TIOWrappedWriter, abridged : bool, terminator : bytes) -> None:
        super().__init__(fobj, abridged)
        self.terminator = terminator

        encoders = tio_default_encoders.copy()
        if abridged:
            encoders[bytes] = _t.cast(_t.Any, self.encode_raw_abridged)
            encoders[str] = _t.cast(_t.Any, self.encode_raw_abridged)
        self.encoder = TIOEncoder(_io.BytesIO(), self.terminator, stdout.encoding,
                                  encoders = encoders, default = self.encode_raw)

    @staticmethod
    def encode_raw(enc : TIOEncoder, obj : _t.Any) -> None:
        if isinstance(obj, EpochMsec):
            enc.write_str(str(int(obj)))
        else:
            raise Failure("can't raw-encode a value of type `%s`", type(obj).__name__)

    @staticmethod
    def encode_raw_abridged(enc : TIOEncoder, obj : _t.AnyStr) -> None:
        abridged, value = abridge_anystr(obj, 256, False)
        if isinstance(obj, bytes):
            enc.write_bytes(value)
        else:
            enc.write_str(value)
        if abridged:
            if isinstance(obj, bytes):
                enc.write_str(f" # {len(obj)} bytes total")
            else:
                enc.write_str(f" # {len(obj)} characters total")

    def emit(self, path : str, names : list[str], values : list[_t.Any]) -> None:
        try:
            for i in range(0, len(values)):
                name, value = names[i], values[i]
                try:
                    self.encoder.encode(value)
                except Failure as exc:
                    exc.elaborate("while encoding attribute `%s'", name)
                    raise exc
                self.encoder.write_bytes(self.terminator)
            self.fobj.write_bytes(self.encoder.fobj.getvalue())
        finally:
            self.encoder.fobj = _io.BytesIO()



# Copyright (c) 2023 Jan Malakhovski <oxij@oxij.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of pwebarc project.

import time as _time
import typing as _t
from kisstdlib.exceptions import *

class CheckedDataClass:
    def __post_init__(self) -> None:
        """__instancecheck__ all fields after __init__"""
        dcf = self.__class__.__dataclass_fields__ # type: ignore
        for k in dcf:
            field = dcf[k]
            v = getattr(self, k)
            typ = field.type
            if not typ.__instancecheck__(v):
                raise TypeError("wrong type while constructing %s: field %s wants %s got %s", type(self), k, typ, type(v))

WrappedValueType = _t.TypeVar("WrappedValueType")
class WrappedValue(_t.Generic[WrappedValueType]):
    def __init__(self, value : WrappedValueType) -> None:
        self.value = value

    def __repr__(self) -> str:
        return "<%s %s>" % (self.__class__.__name__, repr(self.value))

class EpochMsec(WrappedValue[int]):
    @classmethod
    def __instancecheck__(cls, value : _t.Any) -> bool:
        return int.__instancecheck__(value)

    def format(self, fmt : str = "%Y-%m-%d %H:%M:%S") -> str:
        return _time.strftime(fmt, _time.localtime(self.value // 1000)) + "." + format(self.value % 1000, "03")

    def __str__(self) -> str:
        return self.format()

    def __repr__(self) -> str:
        return "<EpochMsec %s>" % (self.format(),)

    def __int__(self) -> int:
        return self.value

def fmt_msec_diff(from_msec : EpochMsec, to_msec : EpochMsec) -> str:
    value = to_msec.value - from_msec.value
    hours = value // 3600000
    value = value % 3600000
    minutes = value // 60000
    value = value % 60000
    seconds = value // 1000
    value = value % 1000
    return str(hours) + ":" + format(minutes, "02") + ":" + format(seconds, "02") + "." + format(value, "03")

def fmt_msec_interval(from_msec : EpochMsec, to_msec : EpochMsec) -> str:
    return f"[{str(from_msec)}]--[{str(to_msec)}] => {fmt_msec_diff(from_msec, to_msec)}"

def rec_get(obj : _t.Any, field : list[str]) -> _t.Any:
    if len(field) == 0:
        return obj

    this, *rest = field
    if isinstance(obj, dict) and this in obj:
        return rec_get(obj[this], rest)
    elif hasattr(obj, "__dataclass_fields__") and this in obj.__dataclass_fields__:
        return rec_get(getattr(obj, this), rest)
    else:
        raise Failure("object of type `%s` does not have an attribute named `%s'", type(obj).__name__, this)

def plainify(obj : _t.Any) -> _t.Any:
    if isinstance(obj, WrappedValue):
        return obj.value
    elif hasattr(obj, "__dataclass_fields__"):
        res = dict()
        for k in obj.__dataclass_fields__:
            res[k] = getattr(obj, k)
        return res
    else:
        raise Failure("can't plainify a value of type `%s`", type(obj).__name__)



# Copyright (c) 2023 Jan Malakhovski <oxij@oxij.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# This file is part of pwebarc project.
#
# This file can be distributed under the terms of the above MIT license or
# Python Software Foundation License version 2 (PSF-2.0) as published by Python
# Software Foundation.

"""LINear State Transformer Domain Specific Language (DSL).

   A caching translator that compiles expressions like "net_url|sha256|prefix 4"
   into Python functions.
"""

import hashlib as _hashlib
import typing as _t
import urllib.parse as _up

from kisstdlib.exceptions import *

# TODO this move somewhere else
def abbrev(v : _t.AnyStr, n : int) -> _t.AnyStr:
    vlen = len(v)
    if vlen > n:
        nn = n // 2
        v = v[:nn] + v[vlen - nn:]
    return v

LinstFunc = _t.Callable[[_t.Any, _t.Any], _t.Any]
LinstAtom = tuple[list[type], _t.Callable[..., LinstFunc]]

class LinstCompileError(Failure): pass
class LinstUnknownAtomError(LinstCompileError): pass

def linst_unknown_atom(name : str) -> LinstAtom:
    raise LinstUnknownAtomError("unknown atom `%s`", name)

def _run_pipe(pipe : list[LinstFunc]) -> LinstFunc:
    def sub(env : _t.Any, v : _t.Any) -> _t.Any:
        for func in pipe:
            v = func(env, v)
        return v
    return sub

_compile_cache : dict[str, LinstFunc] = {}

def linst_compile(expr : str, get_atom : _t.Callable[[str], LinstAtom] = linst_unknown_atom) -> LinstFunc:
    try:
        return _compile_cache[expr]
    except KeyError:
        pass

    pipe = []
    for single in expr.split("|"):
        parts = single.strip().split(" ")
        cmd, *args = parts
        argtypes, func = get_atom(cmd)

        atlen = len(argtypes)
        if atlen != len(args):
            raise LinstCompileError("wrong number of arguments to atom `%s`: expected %d, got %d", cmd, atlen, len(args))

        args_ : list[_t.Any] = []
        for i in range(0, atlen):
            if argtypes[i] is None:
                args_.append(args[i])
            else:
                args_.append(argtypes[i](args[i]))

        func_ = func(*args_)
        pipe.append(func_)

    if len(pipe) == 0:
        raise LinstCompileError("empty pipe")
    elif len(pipe) == 1:
        res = pipe[0]
    else:
        res = _run_pipe(pipe)

    _compile_cache[expr] = res
    return res

def linst_cast(typ : type, arg : _t.Any) -> _t.Any:
    atyp = type(arg)
    if atyp == typ:
        return arg

    if typ is str:
        if atyp is bool or atyp is int or atyp is float:
            return str(arg)
        elif atyp is bytes:
            return arg.decode("utf-8")
    elif atyp is str:
        if typ is bool:
            return bool(arg)
        elif typ is int:
            return int(arg)
        elif typ is float:
            return float(arg)
        elif typ is bytes:
            return arg.encode("utf-8")

    raise AssertionError("can't cast %s to %s", atyp.__name__, typ.__name__)

def linst_cast_val(v : _t.Any, arg : _t.Any) -> _t.Any:
    return linst_cast(type(v), arg)

def linst_const(data : _t.Any) -> LinstAtom:
    def args0() -> _t.Callable[..., LinstFunc]:
        def envfunc(env : _t.Any, v : _t.Any) -> _t.Any:
            return v if v is not None else data
        return envfunc
    return [], args0

def linst_apply0(func : _t.Any) -> LinstAtom:
    def args0() -> _t.Callable[..., LinstFunc]:
        def envfunc(env : _t.Any, v : _t.Any) -> _t.Any:
            return func(v)
        return envfunc
    return [], args0

def linst_apply1(typ : _t.Any, func : _t.Any) -> LinstAtom:
    def args1(arg : _t.Any) -> _t.Callable[..., LinstFunc]:
        def envfunc(env : _t.Any, v : _t.Any) -> _t.Any:
            return func(v, arg)
        return envfunc
    return [typ], args1

def linst_getenv(name : str) -> LinstAtom:
    def args0() -> _t.Callable[..., LinstFunc]:
        def envfunc(env : _t.Any, v : _t.Any) -> _t.Any:
            return v if v is not None else env(name)
        return envfunc
    return [], args0

linst_atoms : dict[str, tuple[str, LinstAtom]] = {
    "es": ('replace `None` value with an empty string `""`',
          linst_const("")),
    "eb": ('replace `None` value with an empty byte string `b""`',
          linst_const(b"")),
    "false": ('replace `None` value with `False`',
          linst_const(False)),
    "true": ('replace `None` value with `True`',
          linst_const(True)),
    "0": ('replace `None` value with `0`',
          linst_const(0)),
    "1": ('replace `None` value with `1`',
          linst_const(1)),
    "not": ("apply logical `not` to value",
          linst_apply0(lambda v: not v)),
    "len": ("apply `len` to value",
          linst_apply0(lambda v: len(v))),
    "str": ("cast value to `str` or fail",
          linst_apply0(lambda v: linst_cast(str, v))),
    "bytes": ("cast value to `bytes` or fail",
          linst_apply0(lambda v: linst_cast(bytes, v))),
    "bool": ("cast value to `bool` or fail",
          linst_apply0(lambda v: linst_cast(bool, v))),
    "int": ("cast value to `int` or fail",
          linst_apply0(lambda v: linst_cast(int, v))),
    "float": ("cast value to `float` or fail",
          linst_apply0(lambda v: linst_cast(float, v))),
    "unquote": ("percent-encoding-unquote value",
          linst_apply0(lambda v: _up.unquote(v))),
    "unquote_plus": ("percent-encoding-unquote value and replace `+` symbols with spaces",
          linst_apply0(lambda v: _up.unquote_plus(v))),
    "sha256": ('compute `hex(sha256(value.encode("utf-8"))`',
          linst_apply0(lambda v: _hashlib.sha256(v if isinstance(v, bytes) else v.encode("utf-8")).hexdigest())),
    "==": ("apply `== arg`, `arg` is cast to the same type as the current value",
          linst_apply1(None, lambda v, arg: v == linst_cast_val(v, arg))),
    "!=": ("apply `!= arg`, similarly",
          linst_apply1(None, lambda v, arg: v != linst_cast_val(v, arg))),
    "<": ("apply `< arg`, similarly",
          linst_apply1(None, lambda v, arg: v < linst_cast_val(v, arg))),
    "<=": ("apply `<= arg`, similarly",
          linst_apply1(None, lambda v, arg: v <= linst_cast_val(v, arg))),
    ">": ("apply `> arg`, similarly",
          linst_apply1(None, lambda v, arg: v > linst_cast_val(v, arg))),
    ">=": ("apply `>= arg`, similarly",
          linst_apply1(None, lambda v, arg: v >= linst_cast_val(v, arg))),
    "prefix": ("take first `arg` characters",
          linst_apply1(int, lambda v, arg: v[:arg])),
    "suffix": ("take last `arg` characters",
          linst_apply1(int, lambda v, arg: v[len(v) - arg:])),
    "abbrev": ("leave the current value as if if its length is less or equal than `arg` characters, otherwise take first `arg/2` followed by last `arg/2` characters",
          linst_apply1(int, abbrev)),
}



# Copyright (c) 2023 Jan Malakhovski <oxij@oxij.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of pwebarc project.

import cbor2 as _cbor2
import dataclasses as _dc
import gzip as _gzip
import io as _io
import logging as _logging
import os as _os
import sys as _sys
import typing as _t
import time as _time
import urllib.parse as _up
import hashlib as _hashlib

from kisstdlib.exceptions import *
from kisstdlib.path import *

from .type import *
from .linst import *

HeadersArray = list[tuple[str, bytes]]
class Headers(WrappedValue[HeadersArray]):
    @classmethod
    def __instancecheck__(cls, value : _t.Any) -> bool:
        return HeadersArray.__instancecheck__(value)

@_dc.dataclass
class Request(CheckedDataClass):
    started_at : EpochMsec
    method : str
    url : str
    headers : Headers
    complete : bool
    body : bytes

    def __post_init__(self) -> None:
        super().__post_init__()
        self.started_at = EpochMsec(_t.cast(int, self.started_at))
        self.headers = Headers(_t.cast(HeadersArray, self.headers))

@_dc.dataclass
class Response(CheckedDataClass):
    started_at : EpochMsec
    code : int
    reason : str
    headers : Headers
    complete : bool
    body : bytes | str

    def __post_init__(self) -> None:
        super().__post_init__()
        self.started_at = EpochMsec(_t.cast(int, self.started_at))
        self.headers = Headers(_t.cast(HeadersArray, self.headers))

@_dc.dataclass
class Reqres(CheckedDataClass):
    version : int
    source : str
    protocol : str
    request : Request
    response : _t.Optional[Response]
    finished_at : EpochMsec
    extra : dict[str, _t.Any]

    def __post_init__(self) -> None:
        super().__post_init__()
        self.finished_at = EpochMsec(self.finished_at) # type: ignore

Reqres_fields = {
    "version": "WEBREQRES format version; int",
    "source": "`+`-separated list of applications that produced this reqres; str",
    "protocol": 'protocol (e.g. `"HTTP/1.0"`, `"HTTP/2.0"`); str',
    "request.started_at": "request start time in milliseconds since 1970-01-01 00:00; EpochMsec",
    "request.method": 'request HTTP method (`"GET"`, `"POST"`, etc); str',
    "request.url": "request URL, including the fragment/hash part; str",
    "request.headers": "request headers; list[tuple[str, bytes]]",
    "request.complete": "is request body complete?; bool",
    "request.body": "request body; bytes",
    "response.started_at": "response start time in milliseconds since 1970-01-01 00:00; EpochMsec",
    "response.code": "HTTP response code (like `200`, `404`, etc); int",
    "response.reason": 'HTTP response reason (like `"OK"`, `"Not Found"`, etc); usually empty for Chromium and filled for Firefox; str',
    "response.headers": "response headers; list[tuple[str, bytes]]",
    "response.complete": "is response body complete?; bool",
    "response.body": "response body; Firefox gives raw bytes, Chromium gives UTF-8 encoded strings; bytes | str",
    "finished_at": "request completion time in milliseconds since 1970-01-01 00:00; EpochMsec"
}

Reqres_derived_attrs = {
    "fs_path": "file system path for the WRR file containing this reqres; str",
    "stime_ms": "aliast for `request.started_at`; int",
    "stime": "`stime_ms` rounded down to seconds (UNIX epoch); int",
    "stime_msq": "three least significant digits of `stime_ms`; int",
    "syear": "year number of `gmtime(rtime)` (UTC year number of `rtime`); int",
    "smonth": "month number of `gmtime(rtime)`; int",
    "sday": "day of the month of `gmtime(rtime)`; int",
    "shour": "hour of `gmtime(rtime)` in 24h format; int",
    "sminute": "minute of `gmtime(rtime)`; int",
    "ssecond": "second of `gmtime(rtime)`; int",

    "rtime_ms": "`response.started_at` if there was a response, `finished_at` otherwise; int",
    "rtime": "`rtime_ms` rounded down to seconds (UNIX epoch); int",
    "rtime_msq": "three least significant digits of `rtime_msq`; int",
    "ryear": "similar to `syear`, but for `rtime`; int",
    "rmonth": "similar to `smonth`, but for `rtime`; int",
    "rday": "similar to `sday`, but for `rtime`; int",
    "rhour": "similar to `shour`, but for `rtime`; int",
    "rminute": "similar to `sminute`, but for `rtime`; int",
    "rsecond": "similar to `ssecond`, but for `rtime`; int",

    "ftime_ms": "aliast for `finished_at`; int",
    "ftime": "`ftime_ms` rounded down to seconds (UNIX epoch); int",
    "ftime_msq": "three least significant digits of `ftime_msq`; int",
    "fyear": "similar to `syear`, but for `ftime`; int",
    "fmonth": "similar to `smonth`, but for `ftime`; int",
    "fday": "similar to `sday`, but for `ftime`; int",
    "fhour": "similar to `shour`, but for `ftime`; int",
    "fminute": "similar to `sminute`, but for `ftime`; int",
    "fsecond": "similar to `ssecond`, but for `ftime`; int",

    "status": '`"NR"` if there was no response, `str(response.code) + "C"` if response was complete, `str(response.code) + "N"` otherwise; str',

    "method": "aliast for `request.method`; str",
    "full_url": "aliast for `request.url`; str",
    "net_url": "`request.url` without the fragment/hash part, if any, this is the URL that actually gets sent to the server; str",
    "scheme": "scheme part of `request.url` (`http`, `https`); str",
    "netloc": "netloc part of `request.url` (i.e., in the most general case, `<username>:<password>@<hostname>:<port>`)",
    "hostname": "hostname part of `request.url`",
    "rhostname": 'hostname part of `request.url` with the order of parts reversed, e.g. `"https://www.example.com"` -> `"com.example.www"`',
    "raw_path": 'raw path part of `request.url`, e.g. `"https://www.example.com"` -> `""`, `"https://www.example.com/"` -> `"/"`, `"https://www.example.com/index.html"` -> `"/index.html"`',
    "path": '`request.url.path_raw` without the leading slash, if any, e.g. `"https://www.example.com"` -> `""`, `"https://www.example.com/"` -> `""`, `"https://www.example.com/index.html"` -> `"index.html"`',
    "ipath": '`path + "index.html"` if `path` is empty or ends with a slash',
    "query": "query part of `request.url` (everything after the `?` character and before the `#` character)",
    "nquery": "normalized `query` (with empty query parameters removed)",
    "oqm": "optional question mark: `?` character if `query` is non-empty and an empty string otherwise",
}

class ReqresExpr:
    reqres : Reqres
    items : dict[str, _t.Any]

    def __init__(self, reqres : Reqres, path : str) -> None:
        self.reqres = reqres
        self.items = {
            "fs_path": path
        }

    def _fill_time(self, prefix : str, ts : int) -> None:
        dt = _time.gmtime(ts)
        self.items[prefix + "year"] = dt.tm_year
        self.items[prefix + "month"] = dt.tm_mon
        self.items[prefix + "day"] = dt.tm_mday
        self.items[prefix + "hour"] = dt.tm_hour
        self.items[prefix + "minute"] = dt.tm_min
        self.items[prefix + "second"] = dt.tm_sec

    def _get_value(self, name : str) -> _t.Any:
        if name.startswith("."):
            name = name[1:]

        try:
            return self.items[name]
        except KeyError:
            pass

        reqres = self.reqres
        if name == "method":
            self.items[name] = reqres.request.method
        elif (name.startswith("s") and \
              name[1:] in ["time", "time_ms", "time_msq", "year", "month", "day", "hour", "minute", "second"]):
            stime_ms = int(reqres.request.started_at)
            stime = stime_ms // 1000
            self.items["stime_ms"] = stime_ms
            self.items["stime"] = stime
            self.items["stime_msq"] = stime_ms % 1000
            self._fill_time("s", stime)
        elif (name.startswith("r") and \
              name[1:] in ["time", "time_ms", "time_msq", "year", "month", "day", "hour", "minute", "second"]) or \
              name == "status":
            if reqres.response is not None:
                rtime_ms = int(reqres.response.started_at)
                status = str(reqres.response.code)
                if reqres.response.complete:
                    status += "C"
                else:
                    status += "N"
            else:
                rtime_ms = int(reqres.finished_at)
                status = "NR"
            self.items["status"] = status
            rtime = rtime_ms // 1000
            self.items["rtime_ms"] = rtime_ms
            self.items["rtime"] = rtime
            self.items["rtime_msq"] = rtime_ms % 1000
            self._fill_time("r", rtime)
        elif (name.startswith("f") and \
              name[1:] in ["time", "time_ms", "time_msq", "year", "month", "day", "hour", "minute", "second"]):
            ftime_ms = int(reqres.finished_at)
            ftime = ftime_ms // 1000
            self.items["ftime_ms"] = ftime_ms
            self.items["ftime"] = ftime
            self.items["ftime_msq"] = ftime_ms % 1000
            self._fill_time("f", ftime)
        elif name in ["full_url", "net_url",
                      "scheme", "netloc", "hostname", "rhostname",
                      "raw_path", "path",
                      "oqm",
                      "query", "nquery"]:
            url = reqres.request.url
            self.items["full_url"] = url
            purl = _up.urlsplit(url)
            query = purl.query
            oqm = "?" if query != "" else ""

            net_url = f"{purl.scheme}://{purl.netloc}{purl.path}{oqm}{query}"
            self.items["net_url"] = net_url

            self.items["scheme"] = purl.scheme
            self.items["netloc"] = purl.netloc

            hostname = purl.hostname
            if hostname is None: assert False
            self.items["hostname"] = hostname

            hparts = hostname.split(".")
            hparts.reverse()
            hostname_rev = ".".join(hparts)
            self.items["rhostname"] = hostname_rev

            self.items["raw_path"] = purl.path
            path = purl.path
            if path.startswith("/"):
                path = path[1:]
            self.items["path"] = path
            if path == "" or path.endswith("/"):
                self.items["ipath"] = path + "index.html"
            else:
                self.items["ipath"] = path
            self.items["oqm"] = oqm
            self.items["query"] = query
            nquery = _up.urlencode(_up.parse_qsl(query))
            self.items["nquery"] = nquery
        elif name == "" or name in Reqres_fields:
            if name == "":
                field = []
            else:
                field = name.split(".")
            # set to None if it does not exist
            try:
                res = rec_get(self.reqres, field)
            except Failure:
                res = None
            self.items[name] = res
        else:
            raise CatastrophicFailure("don't know how to derive `%s`", name)
        return self.items[name]

    def __getattr__(self, name : str) -> _t.Any:
        return self._get_value(name)

    def _get_atom(self, name : str) -> LinstAtom:
        try:
            return linst_atoms[name][1]
        except KeyError:
            return linst_getenv(name)

    def eval(self, expr : str) -> _t.Any:
        ce = linst_compile(expr, self._get_atom)
        return ce(self._get_value, None)

    def __getitem__(self, expr : str) -> _t.Any:
        # this is used in `format_string % self` expressions
        res = self.eval(expr)
        if res is None:
            raise Failure("expression `%s` evaluated to `None`", expr)
        return res

class ParsingError(Failure): pass

def wrr_load(fobj : _io.BufferedReader) -> Reqres:
    head = fobj.peek(2)[:2]
    if head == b"\037\213":
        fobj = _t.cast(_io.BufferedReader, _gzip.GzipFile(fileobj=fobj, mode="rb"))

    try:
        data = _cbor2.load(fobj)
    except _cbor2.CBORDecodeValueError:
        raise ParsingError("can't decode CBOR data, not a CBOR file?")

    if type(data) != list or len(data) == 0:
        raise ParsingError("can't parse CBOR data: wrong structure")

    if data[0] == "WEBREQRES/1":
        _, source, protocol, request_, response_, finished_at, extra = data
        request = Request(*request_)
        if response_ is None:
            response = None
        else:
            response = Response(*response_)

        return Reqres(1, source, protocol, request, response, finished_at, extra)
    else:
        raise ParsingError("can't parse CBOR data: unknown format %s", data[0])

def wrr_loadf(path : str | bytes) -> Reqres:
    with open(path, "rb") as f:
        return wrr_load(f)

def wrr_dump(fobj : _io.BufferedWriter, reqres : Reqres, compress : bool = True) -> None:
    req = reqres.request
    request = req.started_at, req.method, req.url, req.headers.value, req.complete, req.body
    del req

    if reqres.response is None:
        response = None
    else:
        res = reqres.response
        response = res.started_at, res.code, res.reason, res.headers.value, res.complete, res.body
        del res

    structure = ["WEBREQRES/1", reqres.source, reqres.protocol, request, response, reqres.finished_at, reqres.extra]

    if not compress:
        _cbor2.dump(fobj, structure)
        return

    data = _cbor2.dumps(structure)

    # gzip it, if it gzips
    buf = _io.BytesIO()
    with _gzip.GzipFile(fileobj=buf, filename="", mtime=0, mode="wb", compresslevel=9) as gz:
        gz.write(data)
    compressed_data = buf.getvalue()
    del buf

    if len(compressed_data) < len(data):
        data = compressed_data
    del compressed_data

    fobj.write(data)

MapElem = _t.TypeVar("MapElem")
def wrr_map_paths(func : _t.Callable[[Reqres, _t.AnyStr, _t.AnyStr], MapElem],
                  paths : list[_t.AnyStr],
                  errors : str = "fail",
                  follow_symlinks : bool = True) -> _t.Iterable[tuple[_t.AnyStr, MapElem]]:
    for dir_or_file_path in paths:
        dir_or_file_path = _os.path.expanduser(dir_or_file_path)
        for path in walk_orderly(dir_or_file_path,
                                 include_directories = False,
                                 follow_symlinks = follow_symlinks,
                                 handle_error = None if errors == "fail" else _logging.error):
            if (isinstance(path, str) and path.endswith(".part")) or \
               (isinstance(path, bytes) and path.endswith(b".part")):
                continue
            try:
                try:
                    abs_path = _os.path.abspath(path)
                    if not follow_symlinks and _os.path.islink(abs_path):
                        raise Failure("not following a symlink")
                    reqres = wrr_loadf(abs_path)
                except OSError as exc:
                    raise Failure("failed to open")

                res = func(reqres, abs_path, path)
                yield abs_path, res
            except Failure as exc:
                if errors == "ignore":
                    continue
                exc.elaborate("while processing %s", path)
                if errors != "fail":
                    _logging.error("%s", str(exc))
                    continue
                raise exc
